from configargparse import DefaultConfigFileParser


class CustomConfigFileParser(DefaultConfigFileParser):
    def get_syntax_description(self):
        """
        Overwrite string to avoid printing goo.gl URL
        """
        msg = ("Config file syntax allows: key=value, flag=true, stuff=[a,b,c] "
               "(for details, see syntax at https://pypi.org/project/ConfigArgParse/).")
        return msg



r"""
Python API wrapper for the languagetool REST API.

Simple usage::

    >>> from pylanguagetool import api
    >>> api.check(
    ...     'This is a example',
    ...     api_url='https://languagetool.org/api/v2/',
    ...     lang='en-US',
    ... )
    {'software': {'name': 'LanguageTool', 'version': '4.6-SNAPSHOT', 'buildDate': '2019-05-15 19:25', 'apiVersion': 1, 'premium': False, 'premiumHint': 'You might be missing errors only the Premium version can find. Contact us at support<at>languagetoolplus.com.', 'status': ''}, 'warnings': {'incompleteResults': False}, 'language': {'name': 'English (US)', 'code': 'en-US', 'detectedLanguage': {'name': 'English (US)', 'code': 'en-US', 'confidence': 0.561}}, 'matches': [{'message': 'Use "an" instead of \'a\' if the following word starts with a vowel sound, e.g. \'an article\', \'an hour\'', 'shortMessage': 'Wrong article', 'replacements': [{'value': 'an'}], 'offset': 8, 'length': 1, 'context': {'text': 'This is a example', 'offset': 8, 'length': 1}, 'sentence': 'This is a example', 'type': {'typeName': 'Other'}, 'rule': {'id': 'EN_A_VS_AN', 'description': "Use of 'a' vs. 'an'", 'issueType': 'misspelling', 'category': {'id': 'MISC', 'name': 'Miscellaneous'}}, 'ignoreForIncompleteSentence': False, 'contextForSureMatch': 1}]}

"""
import requests


def get_languages(api_url):
    """
    Return supported languages as a list of dictionaries.

    Args:
        api_url (str): API base url.

    Returns:
        List[dict]:
            Supported languages as a list of dictionaries.

            Each dictionary contains three keys, ``name``, ``code`` and
            ``longCode``::

                {
                    "name":"English (GB)",
                    "code":"en",
                    "longCode":"en-GB"
                }

    """
    r = requests.get(api_url + "languages")
    return r.json()


def _is_in_pwl(match, pwl):
    start = match['context']['offset']
    end = start + match['context']['length']
    word = match['context']['text'][start:end]
    return word in pwl


def check(input_text, api_url, lang, mother_tongue=None, preferred_variants=None,
          enabled_rules=None, disabled_rules=None,
          enabled_categories=None, disabled_categories=None,
          enabled_only=False, verbose=False,
          pwl=None,
          **kwargs):
    """
    Check given text and return API response as a dictionary.

    Args:
        input_text (str):
            Plain text that will be checked for spelling mistakes.

        api_url (str):
            API base url, e.g. ``https://languagetool.org/api/v2/``

        lang: Language of the given text as `RFC 3066`__ language code.
            For example ``en-GB`` or ``de-AT``. ``auto`` is a valid value too
            and will cause the language to be detected automatically.

            __ https://www.ietf.org/rfc/rfc3066.txt

        mother_tongue: Native language of the author as `RFC 3066`__ language
            code.

            __ https://www.ietf.org/rfc/rfc3066.txt

        preferred_variants (str):
            Comma-separated list of preferred language variants. The language
            detector used with ``language=auto`` can detect e.g. English, but
            it cannot decide whether British English or American English is
            used. Therefore, this parameter can be used to specify the
            preferred variants like ``en-GB`` and ``de-AT``. Only available
            with ``language=auto``.

        enabled_rules (str):
            Comma-separated list of IDs of rules to be enabled

        disabled_rules (str):
            Comma-separated list of IDs of rules to be disabled.

        enabled_categories (str):
            Comma-separated list of IDs of categories to be enabled.

        disabled_categories (str):
            Comma-separated list of IDs of categories to be disabled.

        enabled_only (bool):
            If ``True``, only the rules and categories whose IDs are specified
            with ``enabledRules`` or ``enabledCategories`` are enabled.
            Defaults to ``False``.

        verbose (bool):
            If ``True``, a more verbose output will be printed. Defaults to
            ``False``.

        pwl (List[str]):
            Personal world list. A custom dictionary of words that should be
            excluded from spell checking errors.

    Returns:
        dict:
            A dictionary representation of the JSON API response.
            The most notable key is ``matches``, which contains a list of all
            spelling mistakes that have been found.

            E.g.::

                {
                    "language": {
                        "code": "en-US",
                        "detectedLanguage": {
                            "code": "en-US",
                            "confidence": 0.561,
                            "name": "English (US)",
                        },
                        "name": "English (US)",
                    },
                    "matches": [
                        {
                            "context": {"length": 1, "offset": 8, "text": "This is a example"},
                            "contextForSureMatch": 1,
                            "ignoreForIncompleteSentence": False,
                            "length": 1,
                            "message": "Use \"an\" instead of 'a' if the following word "
                            "starts with a vowel sound, e.g. 'an article', 'an "
                            "hour'",
                            "offset": 8,
                            "replacements": [{"value": "an"}],
                            "rule": {
                                "category": {"id": "MISC", "name": "Miscellaneous"},
                                "description": "Use of 'a' vs. 'an'",
                                "id": "EN_A_VS_AN",
                                "issueType": "misspelling",
                            },
                            "sentence": "This is a example",
                            "shortMessage": "Wrong article",
                            "type": {"typeName": "Other"},
                        }
                    ],
                    "software": {
                        "apiVersion": 1,
                        "buildDate": "2019-05-15 19:25",
                        "name": "LanguageTool",
                        "premium": False,
                        "premiumHint": "You might be missing errors only the Premium "
                        "version can find. Contact us at "
                        "support<at>languagetoolplus.com.",
                        "status": "",
                        "version": "4.6-SNAPSHOT",
                    },
                    "warnings": {"incompleteResults": False},
                }


    """
    post_parameters = {
        "text": input_text,
        "language": lang,
    }
    if mother_tongue:
        post_parameters["motherTongue"] = mother_tongue
    if preferred_variants:
        post_parameters["preferredVariants"] = preferred_variants
    if enabled_rules:
        post_parameters["enabledRules"] = enabled_rules
    if disabled_rules:
        post_parameters["disabledRules"] = disabled_rules
    if enabled_categories:
        post_parameters["enabledCategories"] = enabled_categories
    if disabled_categories:
        post_parameters["disabledCategories"] = disabled_categories
    if enabled_only:
        post_parameters["enabledOnly"] = 'true'

    r = requests.post(api_url + "check", data=post_parameters)
    if r.status_code != 200:
        raise ValueError(r.text)
    if verbose:
        print(post_parameters)
        print(r.json())
    data = r.json()
    if pwl:
        matches = data.pop('matches', [])
        data['matches'] = [
            match for match in matches
            if not _is_in_pwl(match, pwl)
        ]
    return data



"""
Support spellchecking various file formats by converting them to plain text

"""
import json
import sys
import xml.etree.ElementTree

supported_extensions = ["txt", "html", "md", "markdown", "rst", "ipynb", "json", "xliff"]


def convert(source, texttype):
    """
    Convert files of various types to plaintext

    Args:
        texttype (str):
            file extension of the input file
        source (str):
            content of the input file

    Returns:
        str:
            plaintext output
    """
    if texttype == "html":
        return html2text(source)
    if texttype in ["md", "markdown"]:
        return html2text(markdown2html(source))
    if texttype in ["rst"]:
        return html2text(rst2html(source))
    if texttype == "ipynb":
        return html2text(markdown2html(ipynb2markdown(source)))
    if texttype == "json":
        return transifexjson2txt(source)
    if texttype == "xliff":
        return xliff2txt(source)
    if texttype != "txt":
        print("filetype not detected, assuming plaintext")
    return source


def html2text(html):
    """
    convert HTML to plaintext by parsing it with BeautifulSoup and removing code

    Args:
        html (str):
            HTML string

    Returns:
        str: plaintext
    """
    try:
        from bs4 import BeautifulSoup
    except ImportError:
        notinstalled("beautifulsoup4", "HTML", "text")
        sys.exit(4)
    soup = BeautifulSoup(html, "html.parser")
    # remove scripts from html
    for script in soup(["script", "style", "code", "pre"]) + soup("span", {"class": "literal"}):
        script.extract()
    return soup.get_text()


def markdown2html(markdown):
    """
    convert Markdown to HTML via ``markdown2``

    Args:
        markdown (str):
            Markdown text

    Returns:
        str: HTML
    """
    try:
        import markdown2
    except ImportError:
        notinstalled("markdown2", "markdown", "HTML")
        sys.exit(4)

    return markdown2.markdown(markdown)


def ipynb2markdown(ipynb):
    """
    Extract Markdown cells from iPython Notebook

    Args:
        ipynb (str):
            iPython notebook JSON file

    Returns:
        str: Markdown
    """
    j = json.loads(ipynb)
    markdown = ""
    for cell in j["cells"]:
        if cell["cell_type"] == "markdown":
            markdown += "".join(cell["source"]) + "\n"
    return markdown


def rst2html(rst):
    """
    convert reStructuredText to HTML with ``docutils``

    Args:
        rst (str):
             reStructuredText

    Returns:
        str: HTML
    """
    try:
        from docutils.core import publish_string
    except ImportError:
        notinstalled("docutils", "ReStructuredText", "HTML")
        sys.exit(4)
    return publish_string(rst, writer_name="html5")


def transifexjson2txt(jsondata):
    """
    extract translations from Transifex JSON file

    Args:
        jsondata (str):
            Transifex export file

    Returns:
        str: Plaintext translations

    """
    data = json.loads(jsondata)
    text = ""
    for category, content in data.items():
        for key, value in content.items():
            text += value + "\n"
    return text


def xliff2txt(source):
    """
    extract translations from ``XLIFF`` file

    Args:
        source (str):
            XLIFF XML string

    Returns:
        str: Plaintext translations

    """
    root = xml.etree.ElementTree.fromstring(source)
    text = ""
    for file in root:
        for body in file:
            for trans_unit in body:
                value = trans_unit.find("{urn:oasis:names:tc:xliff:document:1.1}target").text
                if value:
                    text += value + "\n\n"
    return text


def notinstalled(package, convert_from, convert_to):
    print(
        """{package} is needed to convert {source} to {target}
you can install it with pip:
pip install {package}""".format(package=package, source=convert_from, target=convert_to)
    )



"""
A python library and CLI tool for the LanguageTool JSON API.
"""
import os
import sys
from pprint import pprint

import configargparse
import pkg_resources
from colorama import Fore, init as init_colors

from . import CustomConfigFileParser
from . import api
from . import converters

indention = " " * 4


def init_config():
    p = configargparse.ArgParser(default_config_files=["~/.config/pyLanguagetool.conf"],
                                 config_file_parser_class=CustomConfigFileParser.CustomConfigFileParser)
    p.add_argument("-V", "--version", default=False, action='store_true')
    p.add_argument("-v", "--verbose", env_var="VERBOSE", default=False, action='store_true')
    p.add_argument("-a", "--api-url", env_var="API_URL", default="https://languagetool.org/api/v2/",
                   help="the URL of the v2 languagetool API, should end with '/v2/'")
    p.add_argument("--no-color", env_var="NO_COLOR", action='store_true', default=False, help="don't color output")
    p.add_argument("-c", "--clipboard", env_var="CLIPBOARD", action='store_true', default=False,
                   help="get text from system clipboard")
    p.add_argument("-s", "--single-line", env_var="SINGLE_LINE", action='store_true', default=False,
                   help="check every line on its own")
    p.add_argument("-t", "--input-type", env_var="INPUT_TYPE",
                   choices=converters.supported_extensions,
                   help="if not plaintext")
    p.add_argument("-u", "--explain-rule", env_var="EXPLAIN_RULE", action="store_true", default=False,
                   help="print URLs with more information about rules")
    p.add_argument('input file', help='input file', nargs='?')
    p.add_argument("-r", "--rules", env_var="RULES", action='store_true', default=False,
                   help="show the matching rules")
    p.add_argument("--rule-categories", env_var="RULE_CATEGORIES", action='store_true', default=False,
                   help="show the the categories of the matching rules")
    p.add_argument('-l', '--lang', env_var='TEXTLANG', default="auto",
                   help="A language code like en or en-US, or auto to guess the language automatically (see preferredVariants below). For languages with variants (English, German, Portuguese) spell checking will only be activated when you specify the variant, e.g. en-GB instead of just en."
                   )
    p.add_argument("-m", "--mother-tongue", env_var="MOTHER__TONGUE",
                   help="A language code of the user's native language, enabling false friends checks for some language pairs."
                   )
    p.add_argument("-p", "--preferred-variants", env_var="PREFERRED_VARIANTS",
                   help="Comma-separated list of preferred language variants. The language detector used with language=auto can detect e.g. English, but it cannot decide whether British English or American English is used. Thus this parameter can be used to specify the preferred variants like en-GB and de-AT. Only available with language=auto."
                   )
    p.add_argument('-e', '--enabled-rules', env_var='ENABLED_RULES', help='IDs of rules to be enabled, comma-separated')
    p.add_argument('-d', '--disabled-rules', env_var='DISABLED_RULES',
                   help='IDs of rules to be disabled, comma-separated')
    p.add_argument('--enabled-categories', env_var='ENABLED_CATEGORIES',
                   help='IDs of categories to be enabled, comma-separated')
    p.add_argument('--disabled-categories', env_var='DISABLED_CATEGORIES',
                   help='IDs of categories to be disabled, comma-separated')
    p.add_argument("--enabled-only", action='store_true', default=False,
                   help="enable only the rules and categories whose IDs are specified with --enabled-rules or --enabled-categories"
                   )
    p.add_argument(
        '--pwl', '--personal-word-list',
        env_var='PERSONAL_WORD_LIST', help=(
            'File name of personal dictionary. A private dictionary'
            ' can be used to add special words that would otherwise'
            ' be marked as spelling errors.'
        ))

    c = vars(p.parse_args())
    if c["version"]:
        print(pkg_resources.get_distribution('pylanguagetool').version)
        exit()
    if c["enabled_only"] and (c["disabled_categories"] or c["disabled_rules"]):
        print("disabled not allowed")  # TODO: ?
    if c["preferred_variants"] and c["lang"] != "auto":
        # print("You specified --preferred_variants but you didn't specify --language=auto")
        # sys.exit(2)
        print('ignoring --preferred-variants as --lang is not set to "auto"')
        c["preferred_variants"] = None
    if c["verbose"]:
        pprint(c)
    return c, p


def get_clipboard():
    """
    Return text stored in the operating system's clipboard.

    Returns:
        str: Text stored in the operating system's clipboard.

    """
    # See also: http://stackoverflow.com/a/16189232
    import tkinter as tk

    root = tk.Tk()
    root.withdraw()  # keep the window from showing
    clipboard = root.clipboard_get()  # read the clipboard
    root.destroy()
    return clipboard


def get_input_text(config):
    """
    Return text from stdin, clipboard or file.

    Returns:
        Tuple[str, str]:
            A tuple contain of the text and an optional file extension.
            If the text does not come from a file, the extension part of the
            tuple will be none.

    """
    if not sys.stdin.isatty():  # if piped into script
        lines = [line.rstrip() for line in sys.stdin.readlines() if line.rstrip()]
        return "\n".join(lines), None  # read text from pipe and remove empty lines
    elif config["clipboard"]:
        return get_clipboard(), None
    else:
        if config["input file"]:
            extension = os.path.splitext(config["input file"])[1][1:]  # get file extention without .
            try:
                with open(config["input file"], 'r') as myfile:
                    return myfile.read(), extension
            except UnicodeDecodeError:
                print("can't read text")
                sys.exit(1)
        return None, None


def print_errors(response, api_url, print_color=True, rules=False, rule_categories=False, explain_rule=False):
    matches = response["matches"]
    language = response["language"]
    version = response["software"]["name"] + " " + response["software"]["version"]

    def colored(text, color):
        if print_color:
            init_colors()
            return color + text + Fore.RESET
        else:
            return text

    print(colored(
        "{} detected ({:.0f}% confidence)".format(language["detectedLanguage"]["name"],
                                                  language["detectedLanguage"]["confidence"] * 100)
        , Fore.LIGHTBLACK_EX))
    if language["detectedLanguage"]["code"] != language["code"]:
        print(colored(
            "checking as {} text because of setting".format(language["name"])
            , Fore.LIGHTBLACK_EX))
    print()

    tick = colored(u"\u2713", Fore.LIGHTGREEN_EX) + " "
    cross = colored(u"\u2717", Fore.LIGHTRED_EX) + " "

    rule_explanations = []

    for error in matches:
        context_object = error["context"]
        context = context_object["text"]
        length = context_object["length"]
        offset = context_object["offset"]

        endposition = offset + length
        print(error["message"])

        print(
            indention[:2] +
            cross +
            colored(context[:offset], Fore.LIGHTBLACK_EX) +
            colored(context[offset:endposition], Fore.LIGHTRED_EX) +
            colored(context[endposition:], Fore.LIGHTBLACK_EX)
        )
        print(
            indention +
            offset * " " +
            colored(length * "^", Fore.LIGHTRED_EX)
        )

        if error["replacements"]:
            # only print first 5 replacements
            for replacement in error["replacements"][:5]:
                print(
                    indention[:2] +
                    tick +
                    colored(context[:offset], Fore.LIGHTBLACK_EX) +
                    colored(replacement["value"], Fore.LIGHTGREEN_EX) +
                    colored(context[endposition:], Fore.LIGHTBLACK_EX)
                )
        rule = error["rule"]
        if rules:
            print(
                indention[:2] + colored(rule["id"] + ": ", Fore.LIGHTBLACK_EX) + rule["description"]
            )
        if rule_categories:
            category = rule["category"]
            print(
                indention[:2] + colored(category["id"] + ": ", Fore.LIGHTBLACK_EX) + category["name"]
            )
        if explain_rule:
            rule = error["rule"]
            if "description" in rule and "urls" in rule and len(rule["urls"]) > 0:
                rule_explanations.append((rule["description"], rule["urls"][0]["value"]))
        print()

    if explain_rule:
        col_len = max(len(d) for d,u in rule_explanations) + 1
        for descr, url in rule_explanations:
            print(descr + ":" + " " * (col_len - len(descr)) + url)
        print()

    print(colored("Text checked by {url} ({version})".format(url=api_url, version=version), Fore.LIGHTBLACK_EX))


def main():
    config, argparser = init_config()

    if config["verbose"]:
        print(sys.version)

    if config['pwl']:
        with open(config['pwl'], 'r') as fs:
            config['pwl'] = [w.strip() for w in fs.readlines()]

    input_text, inputtype = get_input_text(config)
    if not input_text:
        argparser.print_usage()
        print("input file is required")
        sys.exit(2)
    if config["input_type"]:
        inputtype = config["input_type"]
    if inputtype == "tex":
        print("pyLanguagetool doesn't support LaTeX out of the box.")
        print("But it doesn't have to:")
        print("You can simply use the output of detex")
        if config["input file"]:
            print("    $ detex {} | pylanguagetool".format(config["input file"]))
        print("or use the languagetool integration in TeXstudio.")
        sys.exit(3)
    check_text = converters.convert(input_text, inputtype)
    if config["single_line"]:
        found = False
        for line in check_text.splitlines():
            response = api.check(line, **config)
            print_errors(response,
                         config["api_url"],
                         not config["no_color"],
                         config["rules"],
                         config["rule_categories"]
                         )
            if len(response["matches"]) > 0:
                found = True
        if found:
            sys.exit(1)

    else:
        response = api.check(check_text, **config)
        print_errors(response,
                     config["api_url"],
                     not config["no_color"],
                     config["rules"],
                     config["rule_categories"],
                     config["explain_rule"]
                     )

        if len(response["matches"]) > 0:
            sys.exit(1)



# -*- coding: utf-8 -*-
from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyLanguagetool',
    version='0.9.1',
    packages=find_packages(),
    url='https://pylanguagetool.lw1.at/',
    license='MIT',
    author='Lukas Winkler',
    author_email='python@lw1.at',
    description='A python library and CLI for the LanguageTool JSON API',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Text Processing :: Linguistic'
    ],
    install_requires=['colorama>=0.4.1', 'configargparse>=0.14.0', 'requests>=2.22.0'],
    extras_require={
        'dev': ["pytest"],
        'optional': ["beautifulsoup4", "markdown2", "docutils"],
    },
    keywords="languagetool spell grammar checker",
    entry_points={
        'console_scripts': [
            'pylanguagetool=pylanguagetool:main',
        ],
    },

)
